#!/usr/bin/env node
function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var sade = _interopDefault(require('sade'));
var chalk = _interopDefault(require('chalk'));
var karmatic = _interopDefault(require('./index.js'));
var fs = _interopDefault(require('fs'));
var path = _interopDefault(require('path'));

let {
  write
} = process.stdout;

process.stdout.write = msg => {
  // Strip + prettify console forwarded output:
  let matches = msg.match(/^LOG ([A-Z]+): ([\s\S]*)$/);

  if (matches) {
    msg = chalk.bgBlueBright.white(' ' + matches[1] + ': ') + ' ' + chalk.blue(matches[2]);
  } // Strip browser prefix from the output since there's only one:


  if (msg.match(/^[\n\s]*HeadlessChrome/)) {
    let color = /\bSUCCESS\b/.test(msg) ? 'greenBright' : 'magenta';
    msg = chalk[color](msg.replace(/^[\n\s]*.*?: /g, ''));
  } // Ignore total output since we only have one browser:
  // eslint-disable-next-line no-control-regex


  if (msg.match(/\u001b\[32mTOTAL: /)) return;
  return write.call(process.stdout, msg);
};

let firstFile;
function cleanStack(str, cwd = process.cwd()) {
  str = str.replace(/^[\s\S]+\n\n([A-Za-z]*Error: )/g, '$1');
  firstFile = null;
  let clean = str.replace(new RegExp(`( |\\()(https?:\\/\\/localhost:\\d+\\/base\\/|${cwd.replace(/([\\/[\]()*+$!^.,?])/g, '\\$1')}\\/*)?([^\\s():?]*?)(?:\\?[a-zA-Z0-9]+?)?(:\\d+(?::\\d+)?)`, 'g'), replacer);

  if (firstFile != null) {
    let [filename, line, char] = firstFile;

    if (line) {
      let read;

      try {
        read = fs.readFileSync(path.resolve(cwd, filename), 'utf8');
      } catch (e) {}

      if (read) {
        let start = Math.max(0, char - 40);
        let startLine = Math.max(0, line - 3);
        read = read.split('\n');
        clean = clean.replace(/\n +/g, '\n  ');

        if (line < read.length) {
          clean += '\n\n' + chalk.white(highlight(outdent([read.slice(startLine, line).join('\n'), read[line].substr(start, char + 30), new Array(char).join('-') + '^', read.slice(line + 1, line + 4).join('\n')].join('\n'), '    ', process.stdout.columns - 10), line - startLine, 2));
        }
      }
    }
  }

  return chalk.red(clean);
}

function replacer(str, before, root, filename, position) {
  if (firstFile == null) {
    let [line, char] = position.match(/\d+/g) || [];
    firstFile = [filename, line - 1, char | 0];
  }

  return before + chalk.blue('./' + filename + chalk.dim(position));
}

function highlight(text, line, count) {
  let lines = text.split('\n');
  return chalk.dim(lines.slice(0, line).join('\n')) + '\n' + lines.slice(line, line + count).join('\n') + '\n' + chalk.dim(lines.slice(line + count).join('\n'));
}

function outdent(str, prefix = '', width = 80) {
  str = str.replace(/(^\n+|\n+$)/g, '').replace(/\t/, '    ');
  let indents = str.match(/^[ -]+/gm) || [];
  let minLength = indents.reduce((indent, value) => Math.min(indent, value.length), indents[0] ? indents[0].length : 0);
  str = str.replace(/^[ -]+/gm, str => prefix + str.substring(minLength));
  str = str.replace(/^.*$/gm, str => str.substring(0, width));
  return str;
}

const {
  version
} = require('../package.json');

let toArray = val => typeof val === 'string' ? val.split(/\s*,\s*/) : val == null ? [] : [].concat(val);

let prog = sade('karmatic');
prog.version(version).option('--files', 'Minimatch pattern for test files').option('--headless', 'Run using Chrome Headless', true).option('--coverage', 'Report code coverage of tests', true).option('--downlevel', 'Downlevel syntax to ES5').option('--chromeDataDir', 'Save Chrome preferences');
prog.command('run [...files]', '', {
  default: true
}).describe('Run tests once and exit').action(run);
prog.command('watch [...files]').describe('Run tests on any change').action((str, opts) => run(str, opts, true));
prog.command('debug [...files]').describe('Open a headful Puppeteer instance for debugging your tests').option('--headless', 'Run using Chrome Headless', false) // Override default to false
.option('--browsers', 'Run in specific browsers', null).option('--coverage', 'Report code coverage of tests', false) // Override default to false
.action((str, opts) => run(str, opts, true));
prog.parse(process.argv);

function run(str, opts, isWatch) {
  opts.watch = !!isWatch;
  opts.files = toArray(str || opts.files).concat(opts._);
  const b = opts.browsers || opts.browser;
  opts.browsers = b ? toArray(b) : null;
  karmatic(opts).then(output => {
    if (output != null) process.stdout.write(output + '\n');
    if (!opts.watch) process.exit(0);
  }).catch(err => {
    if (!(typeof err.code === 'number' && err.code >= 0 && err.code < 10)) {
      process.stderr.write(chalk.red(cleanStack(err && (err.stack || err.message) || err)) + '\n');
    }

    process.exit(err.code || 1);
  });
}
//# sourceMappingURL=cli.js.map
