#!/usr/bin/env node
function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var karma = require('karma');
var path = _interopDefault(require('path'));
var puppeteer = _interopDefault(require('puppeteer'));
var chalk = _interopDefault(require('chalk'));
var delve = _interopDefault(require('dlv'));
var fs = _interopDefault(require('fs'));

function fileExists(file) {
  try {
    return fs.statSync(file).isFile();
  } catch (e) {}

  return false;
}
function readFile(file) {
  try {
    return fs.readFileSync(file, 'utf8');
  } catch (e) {}
}
function readDir(file) {
  try {
    return fs.readdirSync(file);
  } catch (e) {}
}
function tryRequire(file) {
  if (fileExists(file)) return require(file);
}
function dedupe(value, index, arr) {
  return arr.indexOf(value) === index;
}
let firstFile;
function cleanStack(str, cwd = process.cwd()) {
  str = str.replace(/^[\s\S]+\n\n([A-Za-z]*Error: )/g, '$1');
  firstFile = null;
  let clean = str.replace(new RegExp(`( |\\()(https?:\\/\\/localhost:\\d+\\/base\\/|${cwd.replace(/([\\/[\]()*+$!^.,?])/g, '\\$1')}\\/*)?([^\\s():?]*?)(?:\\?[a-zA-Z0-9]+?)?(:\\d+(?::\\d+)?)`, 'g'), replacer);

  if (firstFile != null) {
    let [filename, line, char] = firstFile;

    if (line) {
      let read;

      try {
        read = fs.readFileSync(path.resolve(cwd, filename), 'utf8');
      } catch (e) {}

      if (read) {
        let start = Math.max(0, char - 40);
        let startLine = Math.max(0, line - 3);
        read = read.split('\n');
        clean = clean.replace(/\n +/g, '\n  ');

        if (line < read.length) {
          clean += '\n\n' + chalk.white(highlight(outdent([read.slice(startLine, line).join('\n'), read[line].substr(start, char + 30), new Array(char).join('-') + '^', read.slice(line + 1, line + 4).join('\n')].join('\n'), '    ', process.stdout.columns - 10), line - startLine, 2));
        }
      }
    }
  }

  return chalk.red(clean);
}

function replacer(str, before, root, filename, position) {
  if (firstFile == null) {
    let [line, char] = position.match(/\d+/g) || [];
    firstFile = [filename, line - 1, char | 0];
  }

  return before + chalk.blue('./' + filename + chalk.dim(position));
}

function highlight(text, line, count) {
  let lines = text.split('\n');
  return chalk.dim(lines.slice(0, line).join('\n')) + '\n' + lines.slice(line, line + count).join('\n') + '\n' + chalk.dim(lines.slice(line + count).join('\n'));
}

function outdent(str, prefix = '', width = 80) {
  str = str.replace(/(^\n+|\n+$)/g, '').replace(/\t/, '    ');
  let indents = str.match(/^[ -]+/gm) || [];
  let minLength = indents.reduce((indent, value) => Math.min(indent, value.length), indents[0] ? indents[0].length : 0);
  str = str.replace(/^[ -]+/gm, str => prefix + str.substring(minLength));
  str = str.replace(/^.*$/gm, str => str.substring(0, width));
  return str;
}

function babelLoader(options) {
  return {
    test: /\.jsx?$/,
    exclude: /node_modules/,
    loader: require.resolve('babel-loader'),
    query: {
      presets: [[require.resolve('@babel/preset-env'), {
        targets: {
          browsers: ['last 2 Chrome versions', 'last 2 Firefox versions', (options.downlevel || options.browsers && String(options.browsers).match(/(\b|ms|microsoft)(ie|internet.explorer|edge)/gi)) && 'ie>=9'].filter(Boolean)
        },
        corejs: 2,
        useBuiltIns: 'usage',
        modules: false,
        loose: true
      }]],
      plugins: [[require.resolve('@babel/plugin-proposal-object-rest-spread'), {
        loose: true,
        useBuiltIns: true
      }], [require.resolve('@babel/plugin-transform-react-jsx'), {
        pragma: options.pragma || 'h'
      }]].concat(options.coverage ? [require.resolve('babel-plugin-istanbul')] : [])
    }
  };
}

function cssLoader(options) {
  return {
    test: /\.css$/,
    loader: 'style-loader!css-loader'
  };
}

const WEBPACK_VERSION = String(require('webpack').version || '3.0.0');
const WEBPACK_MAJOR = parseInt(WEBPACK_VERSION.split('.')[0], 10);
/**
 * @param {Object} options
 * @param {Array} options.files - Test files to run
 * @param {Array} [options.browsers] - Custom list of browsers to run in
 * @param {Boolean} [options.headless=false] - Run in Headless Chrome?
 * @param {Boolean} [options.watch=false] - Start a continuous test server and retest when files change
 * @param {Boolean} [options.coverage=false] - Instrument and collect code coverage statistics
 * @param {Object} [options.webpackConfig] - Custom webpack configuration
 * @param {Boolean} [options.downlevel=false] - Downlevel/transpile syntax to ES5
 * @param {string} [options.chromeDataDir] - Use a custom Chrome profile directory
 */

function configure(options) {
  let cwd = process.cwd(),
      res = file => path.resolve(cwd, file);

  let files = options.files.filter(Boolean);
  if (!files.length) files = ['**/{*.test.js,*_test.js}'];
  process.env.CHROME_BIN = puppeteer.executablePath();
  let gitignore = (readFile(path.resolve(cwd, '.gitignore')) || '').replace(/(^\s*|\s*$|#.*$)/g, '').split('\n').filter(Boolean);
  let repoRoot = (readDir(cwd) || []).filter(c => c[0] !== '.' && c !== 'node_modules' && gitignore.indexOf(c) === -1);
  let rootFiles = '{' + repoRoot.join(',') + '}';
  const PLUGINS = ['karma-chrome-launcher', 'karma-jasmine', 'karma-spec-reporter', 'karma-sourcemap-loader', 'karma-webpack'].concat(options.coverage ? 'karma-coverage' : []); // Custom launchers to be injected:

  const launchers = {};
  let useSauceLabs = false;
  let browsers;

  if (options.browsers) {
    browsers = options.browsers.map(browser => {
      if (/^chrome([ :-]?headless)?$/i.test(browser)) {
        return `KarmaticChrome${/headless/i.test(browser) ? 'Headless' : ''}`;
      }

      if (/^firefox$/i.test(browser)) {
        PLUGINS.push('karma-firefox-launcher');
        return 'Firefox';
      }

      if (/^sauce-/.test(browser)) {
        if (!useSauceLabs) {
          useSauceLabs = true;
          PLUGINS.push('karma-sauce-launcher');
        }

        const parts = browser.toLowerCase().split('-');
        const name = parts.join('_');
        launchers[name] = {
          base: 'SauceLabs',
          browserName: parts[1].replace(/^(msie|ie|internet ?explorer)$/i, 'Internet Explorer').replace(/^(ms|microsoft|)edge$/i, 'MicrosoftEdge'),
          version: parts[2] || undefined,
          platform: parts[3] ? parts[3].replace(/^win(dows)?[ -]+/gi, 'Windows ').replace(/^(macos|mac ?os ?x|os ?x)[ -]+/gi, 'OS X ') : undefined
        };
        return name;
      }

      return browser;
    });
  } else {
    browsers = [options.headless === false ? 'KarmaticChrome' : 'KarmaticChromeHeadless'];
  }

  if (useSauceLabs) {
    let missing = ['SAUCE_USERNAME', 'SAUCE_ACCESS_KEY'].filter(x => !process.env[x])[0];

    if (missing) {
      throw '\n' + chalk.bold.bgRed.white('Error:') + ' Missing SauceLabs auth configuration.' + '\n  ' + chalk.white(`A SauceLabs browser was requested, but no ${chalk.magentaBright(missing)} environment variable provided.`) + '\n  ' + chalk.white('Try prepending it to your test command:') + '  ' + chalk.greenBright(missing + '=... npm test') + '\n';
    }
  }

  const WEBPACK_CONFIGS = ['webpack.config.babel.js', 'webpack.config.js'];
  let webpackConfig = options.webpackConfig;
  let pkg = tryRequire(res('package.json'));

  if (pkg.scripts) {
    for (let i in pkg.scripts) {
      let script = pkg.scripts[i];

      if (/\bwebpack\b[^&|]*(-c|--config)\b/.test(script)) {
        let matches = script.match(/(?:-c|--config)\s+(?:([^\s])|(["'])(.*?)\2)/);
        let configFile = matches && (matches[1] || matches[2]);
        if (configFile) WEBPACK_CONFIGS.push(configFile);
      }
    }
  }

  if (!webpackConfig) {
    for (let i = WEBPACK_CONFIGS.length; i--;) {
      webpackConfig = tryRequire(res(WEBPACK_CONFIGS[i]));
      if (webpackConfig) break;
    }
  }

  if (typeof webpackConfig === 'function') {
    webpackConfig = webpackConfig({
      karmatic: true
    }, {
      mode: 'development',
      karmatic: true
    });
  }

  webpackConfig = webpackConfig || {};
  let loaders = [].concat(delve(webpackConfig, 'module.loaders') || [], delve(webpackConfig, 'module.rules') || []);

  function evaluateCondition(condition, filename, expected) {
    if (typeof condition === 'function') {
      return condition(filename) == expected;
    } else if (condition instanceof RegExp) {
      return condition.test(filename) == expected;
    }

    if (Array.isArray(condition)) {
      for (let i = 0; i < condition.length; i++) {
        if (evaluateCondition(condition[i], filename)) return expected;
      }
    }

    return !expected;
  }

  function getLoader(predicate) {
    if (typeof predicate === 'string') {
      let filename = predicate;

      predicate = loader => {
        let {
          test,
          include,
          exclude
        } = loader;
        if (exclude && evaluateCondition(exclude, filename, false)) return false;
        if (include && !evaluateCondition(include, filename, true)) return false;
        if (test && evaluateCondition(test, filename, true)) return true;
        return false;
      };
    }

    for (let i = 0; i < loaders.length; i++) {
      if (predicate(loaders[i])) {
        return {
          index: i,
          loader: loaders[i]
        };
      }
    }

    return false;
  }

  function webpackProp(name, value) {
    let configured = delve(webpackConfig, name);

    if (Array.isArray(value)) {
      return value.concat(configured || []).filter(dedupe);
    }

    return Object.assign({}, configured || {}, value);
  }

  const chromeDataDir = options.chromeDataDir ? path.resolve(cwd, options.chromeDataDir) : null;
  const flags = ['--no-sandbox'];
  let generatedConfig = {
    basePath: cwd,
    plugins: PLUGINS.map(req => require.resolve(req)),
    frameworks: ['jasmine'],
    reporters: ['spec'].concat(options.coverage ? 'coverage' : [], useSauceLabs ? 'saucelabs' : []),
    browsers,
    sauceLabs: {
      testName: pkg && pkg.name || undefined
    },
    customLaunchers: Object.assign({
      KarmaticChrome: {
        base: 'Chrome',
        chromeDataDir,
        flags
      },
      KarmaticChromeHeadless: {
        base: 'ChromeHeadless',
        chromeDataDir,
        flags
      }
    }, launchers),
    coverageReporter: {
      reporters: [{
        type: 'text-summary'
      }, {
        type: 'html'
      }, {
        type: 'lcovonly',
        subdir: '.',
        file: 'lcov.info'
      }]
    },

    formatError(msg) {
      try {
        msg = JSON.parse(msg).message;
      } catch (e) {}

      return cleanStack(msg);
    },

    logLevel: 'ERROR',
    loggers: [{
      type: path.resolve(__dirname, 'appender.js')
    }],
    files: [// Inject Jest matchers:
    {
      pattern: path.resolve(__dirname, '../node_modules/expect/build-es5/index.js'),
      watched: false,
      included: true,
      served: true
    }].concat(...files.map(pattern => {
      // Expand '**/xx' patterns but exempt node_modules and gitignored directories
      let matches = pattern.match(/^\*\*\/(.+)$/);
      if (!matches) return {
        pattern,
        watched: true,
        served: true,
        included: true
      };
      return [{
        pattern: rootFiles + '/' + matches[0],
        watched: true,
        served: true,
        included: true
      }, {
        pattern: matches[1],
        watched: true,
        served: true,
        included: true
      }];
    })),
    preprocessors: {
      [rootFiles + '/**/*']: ['webpack'],
      [rootFiles]: ['webpack']
    },
    webpack: {
      devtool: 'cheap-module-eval-source-map',
      // devtool: 'module-source-map',
      mode: webpackConfig.mode || 'development',
      module: {
        // @TODO check webpack version and use loaders VS rules as the key here appropriately:
        rules: loaders.concat(!getLoader(rule => `${rule.use},${rule.loader}`.match(/\bbabel-loader\b/)) ? babelLoader(options) : false
        /*({
        test: /\.[tj]sx?$/,
        // include: files.map(f => minimatch.filter(f, { matchBase: true })),
        exclude: /node_modules/,
        enforce: 'pre',
        loader: require.resolve('istanbul-instrumenter-loader')
        })*/
        , !getLoader('foo.css') && cssLoader()).filter(Boolean)
      },
      resolve: webpackProp('resolve', {
        modules: webpackProp('resolve.modules', ['node_modules', path.resolve(__dirname, '../node_modules')]),
        alias: webpackProp('resolve.alias', {
          [pkg.name]: res('.'),
          src: res('src')
        })
      }),
      resolveLoader: webpackProp('resolveLoader', {
        modules: webpackProp('resolveLoader.modules', ['node_modules', path.resolve(__dirname, '../node_modules')]),
        alias: webpackProp('resolveLoader.alias', {
          [pkg.name]: res('.'),
          src: res('src')
        })
      }),
      plugins: (webpackConfig.plugins || []).filter(plugin => {
        let name = plugin && plugin.constructor.name;
        return /^\s*(UglifyJS|HTML|ExtractText|BabelMinify)(.*Webpack)?Plugin\s*$/gi.test(name);
      }),
      node: webpackProp('node', {}),
      performance: {
        hints: false
      }
    },
    webpackMiddleware: {
      noInfo: true,
      logLevel: 'error',
      stats: 'errors-only'
    },
    colors: true,
    client: {
      captureConsole: true,
      jasmine: {
        random: false
      }
    }
  };

  if (WEBPACK_MAJOR < 4) {
    delete generatedConfig.webpack.mode;
    let {
      rules
    } = generatedConfig.webpack.module;
    delete generatedConfig.webpack.module.rules;
    generatedConfig.webpack.module.loaders = rules;
  }

  return generatedConfig;
}

const karmatic = function (options) {
  try {
    let config = configure(options);
    if (!options.watch) config.singleRun = true;
    let server = createServer(config);
    server.start();
    return Promise.resolve(server.completion);
  } catch (e) {
    return Promise.reject(e);
  }
};

function createServer(config) {
  let resolve, reject;
  let promise = new Promise((res, rej) => {
    resolve = res;
    reject = rej;
  });

  let callback = code => {
    if (code === 0) return resolve();
    let err = Error(`Exit ${code}`);
    err.code = code;
    reject(err);
  };

  let server = new karma.Server(config, callback);
  server.completion = promise;
  return server;
}

module.exports = karmatic;
//# sourceMappingURL=index.js.map
